#include "global.h"
#include<sys/time.h>
#include<stdint.h>
using namespace std;

int32_t nElements;

int _rand()
{
    return rand() % nElements;
}

int main(int argc, char **argv)
{
    int algorithm = atoi(argv[1]);
    nElements = atoi(argv[2]);
    struct timeval startTime, endTime;
    double executedTime;
    
    srand(time(NULL));
    //vetor principal da ordenação
    thrust::device_vector<int> deviceVector(nElements);
    
    cudaDeviceSetLimit(cudaLimitDevRuntimeSyncDepth, MAX_RECURSION);
    if(algorithm == 1)
    {
        cout << " [1] Insertion Sort:" << endl;
        
        thrust::device_vector<int>index(1);
        thrust::host_vector<int> hostVector(nElements);
        
        deviceVector.resize(nElements + 2);
        deviceVector[0] = -99999999;
        deviceVector[nElements + 1] = 99999999;
        
        thrust::generate(hostVector.begin(), hostVector.end(), _rand);
        thrust::copy(hostVector.begin(), hostVector.end(), deviceVector.begin() + 1);
        cudaDeviceSynchronize();
        
        gettimeofday(&startTime, 0);
        insertionSortKernel<<<1, 1>>>(thrust::raw_pointer_cast(&deviceVector[0]),
        nElements, thrust::raw_pointer_cast(&index[0]));
        cudaThreadSynchronize();
        gettimeofday(&endTime, 0);
    }else if(algorithm == 2){
        cout << "[2] Merge Sort:" << endl;
        
        thrust::device_vector<int> A(nElements);
        thrust::device_vector<int> B(nElements);
        
        thrust::sequence(A.begin(), A.end(), 0, 3);
        thrust::sequence(B.begin(), B.end(), 1, 3);
        deviceVector.resize(2 * nElements);
        
        cudaDeviceSynchronize();
        
        gettimeofday(&startTime, 0);
        
        mergeSort(thrust::raw_pointer_cast(&A[0]),
        thrust::raw_pointer_cast(&B[0]),
        thrust::raw_pointer_cast(&deviceVector[0]),
        nElements, nElements);
        cudaThreadSynchronize();
        gettimeofday(&endTime, 0);
        
    }else if(algorithm == 3){
        cout << "[3] Quicksort:" << endl;
        
        thrust::host_vector<int> hostVector(nElements);
        
        thrust::generate(hostVector.begin(), hostVector.end(), _rand);
        thrust::copy(hostVector.begin(), hostVector.end(), deviceVector.begin());
        cudaDeviceSynchronize();
        
        gettimeofday(&startTime, 0);
        
        quickSortKernel<<<1, 1>>>(thrust::raw_pointer_cast(&deviceVector[0]),
        0, nElements - 1);
        cudaThreadSynchronize();
        gettimeofday(&endTime, 0);
    }
    else if(algorithm == 4){
        cout << "[4] Mergesort recursivo:" << endl;
        
        thrust::device_vector<int> tempVector(nElements);
        thrust::host_vector<int> hostVector(nElements);

        thrust::generate(hostVector.begin(), hostVector.end(), _rand);
        thrust::copy(hostVector.begin(), hostVector.end(), deviceVector.begin());
        cudaDeviceSynchronize();

        gettimeofday(&startTime, 0);

        mergeSortKernel<<<1, 1>>>(thrust::raw_pointer_cast(&deviceVector[0]),
                                    thrust::raw_pointer_cast(&tempVector[0]),
                                    0, nElements - 1);
        cudaThreadSynchronize();
        gettimeofday(&endTime, 0);
    }
    else{
        puts("thrust::sort :");
        //float executedTimeThrustSort;
        thrust::host_vector<int> hostVector(nElements);

        thrust::generate(hostVector.begin(), hostVector.end(), _rand);
        thrust::copy(hostVector.begin(), hostVector.end(), deviceVector.begin());
        cudaDeviceSynchronize();

        gettimeofday(&startTime, 0);
        thrust::sort(deviceVector.begin(), deviceVector.end());
        //thrustSort(deviceVector, &executedTimeThrustSort);
        //printf("\nCom thrustSort, a ordenação gastou : %f milisegundos\n", executedTimeThrustSort);
        gettimeofday(&endTime, 0);
    }

    executedTime = endTime.tv_sec + (endTime.tv_usec / 1000000.0);
    executedTime = executedTime - (startTime.tv_sec +(startTime.tv_usec / 1000000.0));
    printf("\n A ordenação gastou : %f segundos\n", executedTime);
    vectorPrinter<<<1, 1>>>(thrust::raw_pointer_cast(&deviceVector[0]), 0, nElements);
    
    cudaDeviceSynchronize();

    return 0;
}

